from typing import Callable, List

import attr
from tensorflow import keras


@attr.s(auto_attribs=True)
class ImageTrainingStep:
    n_epochs: int
    lr: float
    optimizer: keras.optimizers.Optimizer
    freeze: bool = False
    feature_pipeline: List[Callable] = []
    label_pipeline: List[Callable] = []



from pathlib import Path
from typing import Union
from tensorflow import keras


def load_keras_model(
    architecture_path: Union[Path, str], weights_path: Union[Path, str]
):
    with open(architecture_path, "r") as f:
        model = keras.models.model_from_json(f.read())
    model.load_weights(weights_path)
    return model



from typing import Tuple

import pandas as pd
from sklearn.model_selection import train_test_split


def split_df(
    data: pd.DataFrame, test_size: float, target_col: str = None, random_state: int = 42
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    stratify = data[target_col] if target_col else None
    train_data, test_data = train_test_split(
        data, test_size=test_size, stratify=stratify, random_state=random_state
    )
    test_stratify = test_data[target_col] if target_col else None
    val_data, test_data = train_test_split(
        test_data, test_size=0.5, stratify=test_stratify, random_state=random_state
    )
    for df in train_data, val_data, test_data:
        df.reset_index(drop=True, inplace=True)
    return train_data, val_data, test_data



# pylama:ignore=W0611,W0401

import copy
import json
import math
import pickle  # nosec
import re
import shutil
import time
import warnings
from collections import defaultdict, namedtuple
from functools import partial, reduce
from pathlib import Path
from typing import *

import attr
import joblib
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandas_profiling
import requests
import seaborn as sns
import sklearn as sk
from PIL import Image
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.decomposition import PCA, KernelPCA
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.impute import MissingIndicator, SimpleImputer
from sklearn.metrics import (
    accuracy_score,
    classification_report,
    confusion_matrix,
    explained_variance_score,
    f1_score,
    log_loss,
    mean_absolute_error,
    mean_squared_error,
    mean_squared_log_error,
    precision_score,
    r2_score,
    recall_score,
)
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV, train_test_split
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    Binarizer,
    LabelEncoder,
    MinMaxScaler,
    MultiLabelBinarizer,
    Normalizer,
    OneHotEncoder,
    OrdinalEncoder,
    PolynomialFeatures,
    RobustScaler,
    StandardScaler,
)
from sklearn.svm import SVC, SVR, LinearSVC, LinearSVR, OneClassSVM

import lightgbm as lgb
import skopt
from fastprogress import master_bar, progress_bar

try:
    import kaggle
except OSError as error:
    warnings.warn(str(error))



from typing import List, Optional


class DataParams:
    def __init__(
        self,
        target_col: str,
        cat_cols: Optional[List[str]] = None,
        cont_cols: Optional[List[str]] = None,
        text_cols: Optional[List[str]] = None,
        img_cols: Optional[List[str]] = None,
    ):
        self.target_col = target_col
        self.cat_cols = cat_cols or []
        self.cont_cols = cont_cols or []
        self.text_cols = text_cols or []
        self.img_cols = img_cols or []
        self.feature_cols = (
            self.cat_cols + self.cont_cols + self.text_cols + self.img_cols
        )



import os
import shutil
from typing import Optional

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow import keras

from ..data import Dataset
from ..models import load_keras_model, save_keras_model


class ImageLearner:
    def __init__(
        self,
        path,
        base_model,
        input_shape,
        output_shape,
        activation,
        loss,
        metrics,
        dropout=0.0,
        l1=None,
        l2=None,
        override=False,
        load=False,
    ):
        self.path = str(path)
        self.weights_path = f"{self.path}/weights.h5"
        self.architecture_path = f"{self.path}/model.json"
        self.logs_path = f"{self.path}/logs"

        self.input_shape = input_shape
        self.output_shape = output_shape
        self.activation = activation
        self.loss = loss
        self.metrics = metrics
        self.dropout = dropout
        self.l1 = l1
        self.l2 = l2

        self.base_model = base_model(include_top=False, input_shape=input_shape)
        x = keras.layers.concatenate(
            [
                keras.layers.GlobalAvgPool2D()(self.base_model.output),
                keras.layers.GlobalMaxPool2D()(self.base_model.output),
            ]
        )
        x = keras.layers.BatchNormalization()(x)
        x = keras.layers.Dropout(dropout)(x)

        if self.l1 is not None and self.l2 is not None:
            kernel_regularizer = keras.regularizers.l1_l2(self.l1, self.l2)
        elif self.l1 is None and self.l2 is not None:
            kernel_regularizer = keras.regularizers.l2(self.l2)
        elif self.l1 is not None and self.l2 is None:
            kernel_regularizer = keras.regularizers.l1(self.l1)
        else:
            kernel_regularizer = None

        outputs = [
            keras.layers.Dense(
                output_size,
                kernel_regularizer=kernel_regularizer,
                activation=activation,
            )(x)
            for output_size in self.output_shape
        ]

        self.model = keras.Model(inputs=self.base_model.inputs, outputs=outputs)

        if os.path.exists(self.path):
            if load:
                self.load()
            elif override:
                shutil.rmtree(self.path)
                os.makedirs(self.path)
        else:
            os.makedirs(self.path)

        self.save()

    def save(self):
        save_keras_model(self.model, self.architecture_path, self.weights_path)

    def load(self, weights_only: bool = False):
        if weights_only:
            self.model.load_weights(self.weights_path)
        else:
            self.model = load_keras_model(self.architecture_path, self.weights_path)

    def compile(self, optimizer: keras.optimizers.Optimizer, lr: float):
        self.model.compile(
            optimizer=optimizer(lr), loss=self.loss, metrics=self.metrics
        )

    def freeze(self, n_layers: int = 1):
        for layer in self.model.layers[:-n_layers]:
            layer.trainable = False

    def unfreeze(self):
        for layer in self.model.layers:
            layer.trainable = True

    def fit(
        self,
        train_dataset: Dataset,
        validation_dataset: Dataset,
        epochs: int,
        verbose: int = 1,
    ):
        reduce_lr_patience = max(2, epochs // 4)
        early_stopping_patience = reduce_lr_patience * 2

        self.history = self.model.fit(
            x=train_dataset.data,
            steps_per_epoch=train_dataset.steps,
            validation_data=validation_dataset.data,
            validation_steps=validation_dataset.steps,
            epochs=epochs,
            callbacks=[
                keras.callbacks.ModelCheckpoint(
                    self.weights_path, save_best_only=True, save_weights_only=True
                ),
                keras.callbacks.ReduceLROnPlateau(
                    factor=0.3, patience=reduce_lr_patience
                ),
                keras.callbacks.EarlyStopping(
                    patience=early_stopping_patience, restore_best_weights=True
                ),
            ],
            verbose=verbose,
        )
        self.load(weights_only=True)

    def predict(self, path: Optional[str] = None, image=None) -> np.ndarray:
        if image is None:
            image = tf.data.Dataset.from_tensor_slices([path])
            image = self.data.test.preprocess(image, 1).batch(1)
        elif image.ndim == 3:
            image = image[np.newaxis, :]
        return self.model.predict(image)

    def show_history(self, contains: str, skip: int = 0):
        history_df = pd.DataFrame(self.history.history)
        history_df[list(history_df.filter(regex=contains))].iloc[skip:].plot()



import attr

import tensorflow as tf


@attr.s(auto_attribs=True)
class ImageParser:
    n_channels: int = 3

    def __call__(self, filename: str) -> tf.Tensor:
        image = tf.image.decode_jpeg(
            tf.io.read_file(filename), channels=self.n_channels
        )
        image = tf.image.convert_image_dtype(image, tf.float32)

        return image



from sklearn.base import BaseEstimator, TransformerMixin
import pandas as pd
from typing import Dict


class Extractor(BaseEstimator, TransformerMixin):
    def __init__(
        self, source_column: str, patterns: Dict[str, str], drop_source: bool = False
    ):
        self.patterns = patterns
        self.source_column = source_column
        self.drop_source = drop_source

    def fit(self, data: pd.DataFrame) -> "Extractor":
        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data.copy()
        col_index = data.columns.get_loc(self.source_column)
        for i, (column, pattern) in enumerate(self.patterns.items(), 1):
            data.insert(
                col_index + i, column, data[self.source_column].str.extract(pattern)
            )
        if self.drop_source:
            data.drop(self.source_column, axis=1)
        return data



import os
import shutil
from pathlib import Path
from typing import Union

import requests


def download_file(
    url: str, path: Union[Path, str] = Path("."), override: bool = False
) -> str:
    local_filename = f"{str(path)}/{url.split('/')[-1]}"
    if os.path.exists(local_filename) and not override:
        raise FileExistsError(f"File exists: {local_filename}")
    with requests.get(url, stream=True) as r:
        with open(local_filename, "wb") as f:
            shutil.copyfileobj(r.raw, f)

    return local_filename



import os
from typing import List, Optional, Any

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class ImageInputer(BaseEstimator, TransformerMixin):
    def __init__(self, value: Any, columns: Optional[List[str]] = None):
        self.value = value
        self.columns = columns

    def fit(self, data: pd.DataFrame) -> "ImageInputer":
        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        def fill_missing_with_value(value):
            def inner(x):
                return x if os.path.isfile(x) else value

            return inner

        data = data.copy()
        for column in self.columns or data.columns:
            data[column] = (
                data[column].apply(fill_missing_with_value(self.value)).values
            )
        return data



import os
import shutil
from pathlib import Path


def all_files_in_dir(path, extension=None, keep_original=False):
    path = Path(path)
    pathstring = str(path)
    for filename in os.listdir(pathstring):
        if extension and not filename.endswith(extension):
            continue
        shutil.unpack_archive(filename=str(path / filename), extract_dir=pathstring)
        if not keep_original:
            os.remove(str(path / filename))



from typing import Optional, Tuple

import attr

import tensorflow as tf


@attr.s(auto_attribs=True)
class ImageResizer:
    img_dims: Tuple[int, int, int]
    resize: Optional[str] = None
    crop_adjustment: float = 1

    def __call__(self, image: tf.Tensor) -> tf.Tensor:
        height, width, _ = self.img_dims
        if self.resize == "stretch":
            image = tf.image.resize(image, (height, width))
        elif self.resize == "crop":
            crop_height, crop_width = [
                int(x * self.crop_adjustment) for x in (height, width)
            ]
            image = tf.image.resize(
                images=image, size=(crop_height, crop_width), preserve_aspect_ratio=True
            )
            image = tf.image.resize_with_crop_or_pad(image, height, width)
        elif self.resize == "random_crop":
            crop_height, crop_width = [
                int(x * self.crop_adjustment) for x in (height, width)
            ]
            image = tf.image.resize(image, (crop_height, crop_width))
            image = tf.image.random_crop(image, self.img_dims)

        return image



import math
import os
import re
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import sklearn
import tensorflow as tf

from ..data import Dataset
from ..utils import save_file, load_file


class ImageDataset(Dataset):
    @classmethod
    def from_subfolders(cls, path: Union[Path, str]) -> "ImageDataset":
        path = Path(path)
        paths = []
        labels = []
        for label in os.listdir(path):
            for image_path in os.listdir(path / label):
                paths.append(str(path / label / image_path))
                labels.append(label)

        return cls(np.asarray(paths), np.asarray(labels))

    @classmethod
    def from_re(
        cls,
        path: Union[Path, str],
        regex: str,
        default: Optional[Union[int, float, str, bool]] = None,
    ) -> "ImageDataset":
        paths = []
        labels = []
        for value in os.listdir(path):
            match = re.match(regex, value)
            if match:
                labels.append(match.group(1))
            elif default:
                labels.append(default)
            else:
                raise ValueError(
                    f"No match found and no default value provided for value: {value}"
                )
            paths.append(f"{path}/{value}")

        return cls(np.asarray(paths), np.asarray(labels))

    def dataset(
        self,
        batch_size: int,
        img_dims: Tuple[int, int, int],
        shuffle: bool = False,
        prefetch: int = 1,
        n_parallel_calls: int = -1,
    ):
        self.batch_size = batch_size
        self.img_dims = img_dims
        self.shuffle = shuffle
        self.prefetch = prefetch
        self.n_parallel_calls = n_parallel_calls
        self.steps = math.ceil(len(self) / self.batch_size)
        return self

    def make_label_map(self) -> Dict[str, int]:
        return {value: key for key, value in dict(enumerate(np.unique(self.y))).items()}

    def make_label_scaler(self) -> Any:
        label_scaler = sklearn.preprocessing.RobustScaler()
        label_scaler.fit(self.y)
        return label_scaler

    def make_pipeline(
        self,
        regression: bool = False,
        label_map: Optional[Dict[Union[str, int], int]] = None,
        label_scaler: Optional[Any] = None,
        image_pipeline: Optional[List[Callable]] = None,
    ) -> "ImageDataset":
        self.regression = regression
        if self.regression:
            self.label_scaler = self.label_scaler or self.make_label_scaler()
        else:
            self.label_map = label_map or self.make_label_map()
            self.classes = list(self.label_map.keys())
            self.n_classes = len(self.classes)

        self.image_pipeline = image_pipeline or []
        return self

    def load_pipeline(
        self, path: Union[Path, str], regression: bool = False
    ) -> "ImageDataset":
        path = Path(path)
        self.regression = regression
        if self.regression:
            self.label_scaler = load_file(path / "label_scaler.pickle")
        else:
            self.label_map = load_file(path / "label_map.pickle")
            self.classes = list(self.label_map.keys())
            self.n_classes = len(self.classes)

        self.image_pipeline = load_file(path / "image_pipeline.pickle")
        return self

    def save_pipeline(self, path: Union[Path, str]) -> "ImageDataset":
        path = Path(path)
        path.mkdir(parents=True, exist_ok=True)
        if self.regression:
            save_file(self.label_scaler, path / "label_scaler.pickle")
        else:
            save_file(self.label_map, path / "label_map.pickle")

        save_file(self.image_pipeline, path / "image_pipeline.pickle")
        return self

    def preprocess_with_pipeline(
        self, dataset: tf.data.Dataset, pipeline: List[Callable]
    ) -> tf.data.Dataset:
        for fun in pipeline:
            dataset = dataset.map(fun, num_parallel_calls=self.n_parallel_calls)
        return dataset

    def preprocess(self) -> "ImageDataset":
        if self.regression:
            label_ds = tf.data.Dataset.from_tensor_slices(
                self.label_scaler.transform(self.y)
            )
        else:
            label_ds = tf.data.Dataset.from_tensor_slices(
                np.asarray([self.label_map[label] for label in self.y])
            )

        image_ds = tf.data.Dataset.from_tensor_slices(self.x)

        image_ds = self.preprocess_with_pipeline(image_ds, self.image_pipeline)

        dataset = tf.data.Dataset.zip((image_ds, label_ds))
        if self.shuffle:
            dataset = dataset.shuffle(len(self))
        self.data = dataset.repeat().batch(self.batch_size).prefetch(self.prefetch)
        return self

    def show(self, cols: int = 8, n_batches: int = 1):
        if cols >= self.batch_size * n_batches:
            cols = self.batch_size * n_batches
            rows = 1
        else:
            rows = math.ceil(self.batch_size * n_batches / cols)
        _, ax = plt.subplots(rows, cols, figsize=(3 * cols, 3 * rows))
        i = 0
        for x_batch, y_batch in self.data.take(n_batches):
            for (x, y) in zip(x_batch.numpy(), y_batch.numpy()):
                idx = (i // cols, i % cols) if rows > 1 else i % cols
                ax[idx].axis("off")
                ax[idx].imshow(x)
                ax[idx].set_title(y)
                i += 1



from typing import List, Optional

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class CategoricalInputer(BaseEstimator, TransformerMixin):
    def __init__(self, columns: Optional[List[str]] = None, suffix: str = "_na"):
        self.columns = columns
        self.suffix = suffix

    def fit(self, data: pd.DataFrame) -> "CategoricalInputer":
        self.statistics_ = {
            column: data[column].mode().values[0]
            for column in self.columns or data.columns
        }
        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data.copy()
        for column in self.statistics_:
            data.insert(
                data.columns.get_loc(column) + 1,
                column + self.suffix,
                data[column].isna().astype(int),
            )
            data[column].fillna(value=self.statistics_[column], inplace=True)
        return data



from typing import List, Optional

import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin


class CategoricalEncoder(BaseEstimator, TransformerMixin):
    def __init__(self, columns: Optional[List[str]] = None):
        self.columns = columns

    def fit(self, data: pd.DataFrame) -> "CategoricalEncoder":
        data = data.astype("category")
        self.categories_ = {
            column: dict(enumerate(data[column].cat.categories, 1))
            for column in self.columns or data.columns
        }
        return self

    def transform(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data.copy()
        for column in self.categories_:
            data[column] = (
                data[column]
                .map({value: key for key, value in self.categories_[column].items()})
                .fillna(value=0)
                .astype(int)
            )
        return data



import tempfile
from typing import Optional

import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from tensorflow import keras
from typing import List


class LearningRateFinder:
    def __init__(self, model: keras.Model, stop_factor: int = 4, beta: float = 0.98):
        self.model = model
        self.stop_factor = stop_factor
        self.beta = beta
        self.lrs: List[float] = []
        self.losses: List[float] = []
        self.lr_multiplier = 1.0
        self.avg_loss = 0
        self.best_loss = 1e9
        self.batch_num = 0
        self.weights_path: Optional[str] = None

    def reset(self):
        self.lrs = []
        self.losses = []
        self.lr_multiplier = 1.0
        self.avg_loss = 0
        self.best_loss = 1e9
        self.batch_num = 0
        self.weights_path = None

    def on_batch_end(self, batch, logs):
        lr = self.model.optimizer.lr
        self.lrs.append(lr.numpy())
        loss = logs["loss"]
        self.batch_num += 1
        self.avg_loss = (self.beta * self.avg_loss) + ((1 - self.beta) * loss)
        smooth = self.avg_loss / (1 - (self.beta ** self.batch_num))
        self.losses.append(smooth)

        stop_loss = self.stop_factor * self.best_loss

        if self.batch_num > 1 and smooth > stop_loss:
            self.model.stop_training = True
            return

        if self.batch_num == 1 or smooth < self.best_loss:
            self.best_loss = smooth

        lr = lr * self.lr_multiplier
        self.model.optimizer.lr = lr

    def find(
        self,
        x: tf.data.Dataset,
        start_lr: float,
        end_lr: float,
        epochs: Optional[int] = None,
        steps_per_epoch: int = 1,
        batch_size: int = 32,
        sample_size: int = 2048,
        verbose: int = 1,
    ):
        self.reset()
        if epochs is None:
            epochs = int(np.ceil(sample_size / steps_per_epoch))

        num_batch_updates = epochs * steps_per_epoch

        self.lr_multiplier = (end_lr / start_lr) ** (1.0 / num_batch_updates)

        self.weights_path = tempfile.mkstemp()[1]
        self.model.save_weights(self.weights_path)

        original_lr = self.model.optimizer.lr.numpy()
        self.model.optimizer.lr = start_lr

        self.model.fit(
            x,
            epochs=epochs,
            steps_per_epoch=steps_per_epoch,
            callbacks=[keras.callbacks.LambdaCallback(on_batch_end=self.on_batch_end)],
            verbose=verbose,
        )

        self.model.load_weights(self.weights_path)
        self.model.optimizer.lr = original_lr

    def plot_loss(
        self, skip_begin: int = 10, skip_end: int = 1, title: Optional[str] = None
    ):
        lrs = self.lrs[skip_begin:-skip_end]
        losses = self.losses[skip_begin:-skip_end]

        plt.plot(lrs, losses)
        plt.xscale("log")
        plt.xlabel("Learning Rate (Log Scale)")
        plt.ylabel("Loss")

        if title:
            plt.title(title)



import math
from typing import List, Optional, Tuple, Union

import numpy as np
import pandas as pd


class Dataset:
    def __init__(self, x: np.ndarray, y: np.ndarray):
        self.x = x
        self.y = y

    @classmethod
    def split(
        cls,
        dataset: "Dataset",
        fracs: Union[List[float], Tuple[float]],
        random: Optional[bool] = True,
    ) -> Tuple["Dataset"]:
        x = dataset.x
        y = dataset.y

        if random:
            random_indices = np.random.permutation(len(dataset))
            x = x[random_indices]
            y = y[random_indices]

        result = []
        current_index = 0
        for frac in fracs:
            dx = math.ceil(len(dataset) * frac)
            split_dataset = cls(
                x=x[current_index : current_index + dx],
                y=y[current_index : current_index + dx],
            )
            result.append(split_dataset)
            current_index += dx

        return tuple(result)

    @classmethod
    def from_dataframe(
        cls, dataframe: pd.DataFrame, x_col: str, y_col: str
    ) -> "Dataset":
        return cls(dataframe[x_col].values, dataframe[y_col].values)

    def __len__(self) -> int:
        return len(self.y)



import os
from ast import literal_eval

from setuptools import find_packages, setup


def get_version(version_tuple):
    if not isinstance(version_tuple[-1], int):
        return ".".join(map(str, version_tuple[:-1])) + version_tuple[-1]
    return ".".join(map(str, version_tuple))


init = os.path.join(os.path.dirname(__file__), "toai", "__init__.py")

version_line = list(filter(lambda l: l.startswith("VERSION"), open(init)))[0]

# VERSION is a tuple so we need to eval 'version_line'.
# We could simply import it from the package but we
# cannot be sure that this package is importable before
# installation is done.
PKG_VERSION = get_version(literal_eval(version_line.split(" = ")[-1]))

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="toai",
    version=PKG_VERSION,
    description="To AI helper library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dovydas Ceilutka",
    author_email="dovydas.ceilutka@gmail.com",
    url="https://github.com/mokahaiku/toai",
    download_url="https://github.com/mokahaiku/toai",
    license="MIT",
    install_requires=[
        "attrs",
        "fastparquet",
        "fastprogress",
        "funcy",
        "imagehash",
        "joblib",
        "kaggle",
        "lightgbm",
        "numpy",
        "pandas-profiling",
        "pandas",
        "Pillow",
        "pyarrow",
        "scikit-image",
        "scikit-learn",
        "scikit-optimize",
        "seaborn",
    ],
    extras_require={
        "tests": [
            "pytest",
            "pytest-pep8",
            "pytest-xdist",
            "pytest-cov",
            "pytest-timeout",
        ]
    },
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
    packages=find_packages(),
)



import time
from typing import List

import attr
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf

from .ImageTrainingStep import ImageTrainingStep
from .ImageLearner import ImageLearner
from ..data import DataContainer


@attr.s(auto_attribs=True)
class ImageTrainer:
    learner: ImageLearner
    data_container: DataContainer
    steps: List[ImageTrainingStep]
    template: str = "Name: {} Train Time: {:.1f} min. Eval Time: {:.2f}s Loss: {:.4f} Accuracy: {:.2%}"

    def train(self):
        start_time = time.time()
        for step in self.steps:
            self.learner.freeze() if step.freeze else self.learner.unfreeze()
            self.learner.compile(optimizer=step.optimizer, lr=step.lr)
            self.data_container.train.image_pipeline = step.feature_pipeline
            self.data_container.train.preprocess()
            self.learner.fit(
                self.data_container.train, self.data_container.validation, step.n_epochs
            )
        end_time = time.time()

        eval_start_time = time.time()
        evaluation_results = self.evaluate_dataset(verbose=0)
        eval_end_time = time.time()

        print("-".center(80, "-"))
        print(
            self.template.format(
                self.learner.base_model.name,
                (end_time - start_time) / 60,
                (eval_end_time - eval_start_time),
                *evaluation_results,
            )
        )
        print("-".center(80, "-"))

    def evaluate_dataset(self, mode: str = "validation", verbose: int = 1):
        dataset = getattr(self.data_container, mode)
        return self.learner.model.evaluate(
            dataset.data, steps=dataset.steps, verbose=verbose
        )

    def predict_dataset(self, mode: str = "validation", verbose: int = 0):
        dataset = getattr(self.data_container, mode)
        return self.learner.model.predict(
            dataset.data, steps=dataset.steps, verbose=verbose
        )

    def analyse_dataset(self, mode: str = "validation", verbose: int = 0):
        dataset = getattr(self.data_container, mode)
        image_ds = tf.data.Dataset.from_tensor_slices(dataset.x)
        image_ds = dataset.preprocess_with_pipeline(
            image_ds, dataset.image_pipeline
        ).batch(1)
        images = [
            img[0].numpy() for img in image_ds.take(dataset.steps * dataset.batch_size)
        ]
        probs = self.learner.model.predict(image_ds)
        pred_code = probs.argmax(axis=1)
        label_code = [dataset.label_map[label] for label in dataset.y]
        inverse_label_map = {value: key for key, value in dataset.label_map.items()}
        pred = [inverse_label_map[x] for x in pred_code]
        return pd.DataFrame.from_dict(
            {
                "path": dataset.x,
                "image": images,
                "label": dataset.y,
                "label_code": label_code,
                "pred": pred,
                "pred_code": pred_code,
                "label_probs": probs[:, label_code][np.eye(len(dataset.y), dtype=bool)],
                "pred_probs": probs[:, pred_code][np.eye(len(pred_code), dtype=bool)],
            }
        )

    def show_predictions(
        self,
        mode: str = "validation",
        correct: bool = False,
        ascending: bool = True,
        cols: int = 8,
        rows: int = 2,
    ):
        df = self.analyse_dataset(mode=mode)
        df = df[(df.label == df.pred) if correct else (df.label != df.pred)]
        df.sort_values(by=["label_probs"], ascending=ascending, inplace=True)
        _, ax = plt.subplots(rows, cols, figsize=(3 * cols, 3.5 * rows))
        for i, row in enumerate(df.head(cols * rows).itertuples()):
            idx = (i // cols, i % cols) if rows > 1 else i % cols
            ax[idx].axis("off")
            ax[idx].imshow(row.image)
            ax[idx].set_title(
                f"{row.label}:{row.pred}\n{row.label_probs:.4f}:{row.pred_probs:.4f}"
            )



from pathlib import Path
from typing import Union

from tensorflow import keras


def save_keras_model(
    model: keras.Model,
    architecture_path: Union[Path, str],
    weights_path: Union[Path, str],
):
    model.save_weights(weights_path)
    with open(architecture_path, "w") as f:
        f.write(model.to_json())



# pylama:ignore=W0611

from .ImageAugmentor import ImageAugmentor
from .ImageDataset import ImageDataset
from .ImageLearner import ImageLearner
from .ImageParser import ImageParser
from .ImageResizer import ImageResizer
from .ImageTrainer import ImageTrainer
from .ImageTrainingStep import ImageTrainingStep
from .LearningRateFinder import LearningRateFinder



import attr
from typing import Optional
import tensorflow as tf


@attr.s(auto_attribs=True)
class ImageAugmentor:
    level: int = 0
    flips: Optional[str] = None

    def __call__(self, image: tf.Tensor) -> tf.Tensor:
        if self.flips in ["horizontal", "both"]:
            image = tf.image.random_flip_left_right(image)
        if self.flips in ["vertical", "both"]:
            image = tf.image.random_flip_up_down(image)

        if self.level > 0:
            lower = 1 - 0.1 * self.level
            upper = 1 + 0.1 * self.level
            min_jpeg_quality = max(0, int((lower - 0.5) * 100))
            max_jpeg_quality = min(100, int((upper - 0.5) * 100))
            image = tf.image.random_jpeg_quality(
                image,
                min_jpeg_quality=min_jpeg_quality,
                max_jpeg_quality=max_jpeg_quality,
            )
            image = tf.image.random_contrast(image, lower=lower, upper=upper)
            image = tf.image.random_saturation(image, lower=lower, upper=upper)
            image = tf.clip_by_value(image, 0.0, 1.0)
        return image
